import hashlib
import secrets

from ..exceptions import *

def _parse_token(s):
	SEP = [b'(', b')', b'<', b'>', b'@', b',', b';', b':', b'\\', b'\'', b'/',
	       b'[', b']', b'?', b'=', b'{', b'}', b' ', b'\t']
	CTL = [bytes([c]) for c in range(0, 31)] + [b'127']
	for index in range(len(s)):
		c = bytes([s[index]])
		if c in SEP + CTL:
			return bytes(s[:index]), bytes(s[index:])
	return s, b''

def _parse_qstr(s):
	if s[0] != b'"'[0]:
		raise ValueError()
	res = b''
	escaped = False
	for index in range(1, len(s)):
		c = bytes([s[index]])
		if escaped:
			res += c
			escaped = False
		elif c == b'\\':
			escaped = True
		elif c == b'"':
			return res, bytes(s[index+1:])
		else:
			res += c
	raise ValueError()

def _parse_token_qstr(s):
	if s[0] == b'"'[0]:
		return _parse_qstr(s)
	return _parse_token(s)

def _parse_kwargs(s):
	res = []
	while True:
		key, s = _parse_token(s)
		if s[0] != b'='[0]:
			raise ValueError()
		value, s = _parse_token_qstr(bytes(s[1:]))
		res.append((key, value))
		if not s:
			return res
		if s[0] != b','[0]:
			raise ValueError()
		s = bytes(s[1:])
	return res

def _generate_nonce():
	return secrets.token_urlsafe(1024).encode()

def _hexdigest(data):
	ctx = hashlib.md5()
	ctx.update(data)
	return ctx.hexdigest().lower().encode()

def _handle_ldap_bind(get_credentials, get_nonce=_generate_nonce, initial_response=None):
	# Defined by RFC2831 and RFC2829, obsoleted by RFC6331
	nonce = get_nonce()
	challenge = b'nonce="%s",charset="utf-8",algorithm="md5-sess"'%(nonce)
	resp = yield challenge
	args = {key: value for key, value in _parse_kwargs(resp)}
	if args[b'nonce'] != nonce:
		raise LDAPProtocolError()
	try:
		charset = 'utf-8' if args.get(b'charset', b'utf-8') == b'utf-8' else 'latin_1'
		username = args[b'username']
		realm = args.get(b'realm', b'')
		cnonce = args[b'cnonce']
		nc = args.get(b'nc', b'00000001')
		qop = args.get(b'qop', b'auth')
		digest_uri = args[b'digest-uri']
		parts = digest_uri.decode(charset).split('/', 2)
		serv_type = parts[0]
		host = parts[1]
		serv_name = parts[2] if len(parts) == 3 else None
		authzid = args[b'authzid'].decode(charset) if b'authzid' in args else None
		response = args[b'response']
	except (KeyError, IndexError):
		raise LDAPProtocolError()
	except UnicodeError:
		raise LDAPProtocolError()
	if serv_type != 'ldap':
		raise LDAPInvalidCredentials()
	valid_credentials = get_credentials(username, realm, host, serv_name=None, authzid=None, charset=charset)
	a2 = b'AUTHENTICATE:' + digest_uri
	data = nonce + b':' + nc + b':' + cnonce + b':' + qop + b':' + _hexdigest(a2)
	for bind_obj, pwdigest in valid_credentials:
		a1 = pwdigest + b':' + nonce + b':' + cnonce
		key = _hexdigest(a1)
		expected_response = _hexdigest(key + b':' + data)
		if expected_response != response:
			continue
		# We don't support subsequent authentication so according to RFC 2829 the
		# serverSaslCreds field in our response should be absent and we should
		# return (bind_obj, None). But this seems to confuse some clients (e.g.
		# openldap's ldapsearch) so we return serverSaslCreds with rspauth instead.
		a2 = b':' + digest_uri
		data = nonce + b':' + nc + b':' + cnonce + b':' + qop + b':' + _hexdigest(a2)
		response = b'rspauth=%s'%_hexdigest(key + b':' + data)
		return bind_obj, response
	raise LDAPInvalidCredentials()

def _encode_latin1_or_utf8(s):
	try:
		return s.encode('latin_1')
	except UnicodeEncodeError:
		pass
	return s.encode('utf-8')

def credential_digest(username, realm, password):
	'''Compute DIGEST-MD5-specific credential digest

	:param username: Name of the user account
	:type username: bytes or str
	:param realm: Realm containing the user account
	:type realm: bytes or str
	:param password: Password for the user account
	:type password: bytes or str

	:returns: DIGEST-MD5-specific credential digest (16 bytes)
	:rtype: bytes

	Parameters passed as strings are encoded according to the special DIGEST-MD5
	encoding rules (latin_1 whenever all characters can be encoded wit it, utf-8
	otherwise).'''
	if isinstance(username, str):
		username = _encode_latin1_or_utf8(username)
	if isinstance(realm, str):
		realm = _encode_latin1_or_utf8(realm)
	if isinstance(password, str):
		password = _encode_latin1_or_utf8(password)
	ctx = hashlib.md5()
	ctx.update(username + b':' + realm + b':' + password)
	return ctx.digest()
